'use strict';

var _ = require('underscore');

var modules = {
  server: {}, //name -> package.json
  shared: {},
  client: {}
};

var inverseClientDeps = {}; //depName -> name -> true

var addClientDeps = function(desc) {
  //TODO: do I not need to track client->shared deps here?
  var name = desc.name;
  var clientDeps = desc.modulo.clientDependencies;
  _.each(desc.modulo.clientDependencies, function(depName) {
    if (!_.has(inverseClientDeps, depName)) { inverseClientDeps[depName] = {}; }
    inverseClientDeps[depName][name] = true;
  });
};

var removeClientDeps = function(desc) {
  if (!desc) { return; }

  var name = desc.name;
  _.each(desc.modulo.clientDependencies, function(depName) {
    delete inverseClientDeps[depName][name];
    if (_.isEmpty(inverseClientDeps[depName])) { delete inverseClientDeps[depName]; }
  });
};

var load = function(parsed) {
  if (parsed.modulo.group === 'client') {
    removeClientDeps(modules[parsed.modulo.group][parsed.name]);
    addClientDeps(parsed);
  }

  modules[parsed.modulo.group][parsed.name] = parsed;
};

var expandDeps = function(group, deps) {
  //TODO: some caching on this (or pre-calculate on load)
  var pending = { shared: {} };
  var done = { shared: {} };
  var result = { shared: [] };
  pending[group] = _.object(deps, []);
  done[group] = {};
  result[group] = [];

  var expandNode = function(nodeGroup, name) {
    delete pending[nodeGroup][name];

    if (!_.has(modules[nodeGroup], name) || _.has(done[nodeGroup], name)) {
      return;
    }

    done[nodeGroup][name] = true;
    if (modules[nodeGroup][name].modulo.type !== 'meta') {
      result[nodeGroup].push(name);
    }

    var addDeps = function(depList, depGroup) {
      _.each(depList, function(depName) {
        if (_.has(modules[depGroup], depName) && !_.has(done[depGroup], depName)) {
          pending[depGroup][depName] = true;
        }
      });
    };

    //TODO: deal with virtual modules or replacement modules

    var nodeDesc = modules[nodeGroup][name].modulo;

    addDeps(nodeDesc.sharedDependencies, 'shared');
    if (nodeGroup === 'server') {
      addDeps(nodeDesc.serverDependencies, 'server');
    } else if (nodeGroup === 'client') {
      addDeps(nodeDesc.clientDependencies, 'client');
    }
  };

  while(!_.isEmpty(pending[group]) || !_.isEmpty(pending.shared)) {
    _.each(pending[group], function(ignored, name) { expandNode(group, name); });
    _.each(pending.shared, function(ignored, name) { expandNode('shared', name); });
  }

  var ret = { shared: result.shared.sort() };
  ret[group] = result[group].sort();
  return ret;
};

var filterClient = function(serverMods) {
  var serverModMap = _.object(serverMods, []);

  var optional = _.object(_.keys(modules.client), []);
  var unsupported = {};
  var recommended = {};

  var addRecommendedClientMod = function(clientModName) {
    if (_.has(modules.client, clientModName)) {
      delete optional[clientModName];
      recommended[clientModName] = true;
    }
  };

  _.each(serverMods, function(serverModName) {
    _.each(modules.server[serverModName].modulo.recommendedClientMods, addRecommendedClientMod);
  });

  //TODO: any client mod that depends on a client mod that depends on a missing server dep should also be unsupported
  var missingServerDependency = function(serverModName) {
    return !_.has(serverModMap, serverModName);
  };

  var pendingUnsupported = {};
  _.each(modules.client, function(ignored, clientModName) {
    if (_.any(modules.client[clientModName].modulo.requiredServerMods, missingServerDependency)) {
      pendingUnsupported[clientModName] = true;
    }
  });

  var expandNode = function(ignored, nodeName) {
    delete pendingUnsupported[nodeName];
    delete optional[nodeName];
    delete recommended[nodeName];
    if (_.has(modules.client, nodeName)) {
      unsupported[nodeName] = true;
      if (_.has(inverseClientDeps, nodeName)) {
        _.each(inverseClientDeps[nodeName], function(ignored, parentName) {
          if(_.has(modules.client, parentName) && !_.has(unsupported, parentName)) {
            pendingUnsupported[parentName] = true;
          }
        });
      }
    }
  };

  while(!_.isEmpty(pendingUnsupported)) {
    _.each(pendingUnsupported, expandNode);
  }

  return {
    recommended: _.keys(recommended).sort(),
    optional: _.keys(optional).sort(),  //TODO: should I not list "optional" since that includes all mods not otherwise enumerated?
    unsupported: _.keys(unsupported).sort()
  };
};

exports.load = load;
exports.expandServer = function(deps) { return expandDeps('server', deps); };
exports.expandClient = function(deps) { return expandDeps('client', deps); };
exports.filterClient = filterClient;
